#!/usr/bin/perl -w

package Test::Repocop::Fixscripts;

use strict;
use warnings;
use Carp;
use Getopt::Long;
use RPM::Source::Editor;
use Test::Repocop::Common;
use File::Basename;
use File::Path;
use File::Temp qw/tempfile tempdir/;

# our -- hack ?
our $alttoolsdir=dirname($0);
# for cross-references in testnames

my @fixscriptdirs = ('/usr/share/repocop/fixscripts');
push @fixscriptdirs, $alttoolsdir.'/fixscripts' if -d $alttoolsdir.'/fixscripts';
push @INC, @fixscriptdirs;

our %fixscriptmap;
our @SPECHOOKS;

my @postcleanscripts;
foreach my $dir (@fixscriptdirs) {
    push @postcleanscripts, map {basename($_)} map {s!.pl$!!;$_} glob ("$dir/postclean-*.pl");
}

#print STDERR 'postclean scripts: ',join(', ',@postcleanscripts),"\n";
&__preload_pl(1, @postcleanscripts);

# public static members
sub patchname2srcid {
    my $id=basename(shift());
    $id=~s/(?:-(?:digest|cummulative|safe))?\.(?:patch|diff)$//;
    $id.='.src' if $id!~/\.src$/;
    return $id;
}

# public members
sub new {
    my $class=shift;
    my $self = {
	FIX_PKGLIST => [],
	FIX_TESTNAMELIST => [],
	VERBOSE => 0,
	# TODO! support it! not yet supported :(
	FORCE_PACKAGER => undef,
	PATCH_AUTHOR_NAME => 'Repocop Q. A. Robot',
	PATCH_AUTHOR_EMAIL=> 'repocop@altlinux.org',
	DEF_CHANGELOG => '- NMU (by repocop): the following fixes applied:'."\n",
	_SPEC_CHANGED => 0,
	@_
    };

    if (@{$self->{FIX_PKGLIST}}==0 || @{$self->{FIX_TESTNAMELIST}}==0) {
	die "ERROR: FIX_PKGLIST and FIX_TESTNAMELIST are required!\n";
    }

    bless $self, $class;
    $self->__preload_fixscripts();
    return $self;
}

sub _get_default_changelog {
    my $self=shift;
    return $self->{DEF_CHANGELOG};
}

sub __preload_fixscripts {
    my $self=shift;
    my %unique_testnames=map {$_=>1} @{$self->{FIX_TESTNAMELIST}};
    &__preload_pl($self->{VERBOSE}, keys %unique_testnames);
}

sub __preload_pl {
    my $verbose=shift;
    foreach my $testname (@_) {
	unless ($fixscriptmap{$testname}) {
	    eval {
		require $testname.'.pl';
	    };
	    warn "can't load fixscript for $testname: skipped\n" if $@ and $verbose;
	    $fixscriptmap{$testname}=$SPECHOOKS[0] if @SPECHOOKS;
	    $#{SPECHOOKS}=-1;
	}
    }
}


sub _call_fixscript {
    my $self=shift;
    my ($testname,$pkg)=@_;
    #print STDERR "applying $testname for $pkg\n" if $self->{VERBOSE}>1;
    my $fixscriptref=$fixscriptmap{$testname};
    $pkg||='spec file';
    if ($fixscriptref) {
	&Test::Repocop::Common::set_test_environment($testname);
#	$self->{EDITOR}->{WARN_NOTAPPLIED}++;
	&{$fixscriptref}($self->{EDITOR},$pkg);
#	$self->{EDITOR}->{WARN_NOTAPPLIED}--;
	$self->{DEF_CHANGELOG}.="  * $testname for $pkg\n";
	$self->{__CALL_FIXSCRIPT}=1;
	print STDERR "applied $testname for $pkg\n" if $self->{VERBOSE};
    } else {
	print STDERR "no fixscript for $testname for $pkg\n" if $self->{VERBOSE};
    }
}

sub apply_fixscripts {
    my $self=shift;
    my $opinfo = {
	@_
    };
    my $spec=$self->{EDITOR};
    my $before=$spec->get_spec();

    my @testnames=@{$self->{FIX_TESTNAMELIST}};
    my $i=0;
    for ($i=0;$i<@testnames;$i++) {
	$self->_call_fixscript($testnames[$i],$self->{FIX_PKGLIST}->[$i]);
    }
    # postclean scripts
    map {$self->_call_fixscript($_)} @postcleanscripts;

    if ($before ne $spec->get_spec()) {
	$self->{__SPEC_CHANGED}=1;
	return 1;
    } else {
	return 0;
    }
}

sub __increment_release {
    my ($self,$altrel,$policy)=@_;
    $policy||= 'qa';
    my $newrel;
    if ($policy eq 'qa') {
	if ($altrel=~s/\.qa(\d+)(?:\.(?:qa\d+|1))*$//) {
	    $newrel = $altrel.".qa".($1+1);
	} else {
	    $newrel = $altrel.".qa1";
	}
    } elsif ($policy eq 'dd') {
	if ($altrel=~s/\.(\d+)$//) {
	    $newrel = $altrel.".".($1+1);
	} else {
	    $newrel = $altrel.".1";
	}
    } elsif ($policy eq 'do') {
	$newrel = $altrel.".1";
    } else {
	die "Unknown increment release policy: $policy";
    }
    return $newrel;
}

sub raise_release {
    my $self=shift;
    my $relinfo = {
	@_
    };
    $relinfo->{CHANGELOG} ||= $self->{DEF_CHANGELOG};

    my $spec=$self->{EDITOR};
    my $old_evr=$spec->get_evr();
    
    if ($relinfo->{RELEASE}) {
	$spec->get_section('package','')->set_tag('Release',$relinfo->{RELEASE});
    } else {
	my $mainsec=$spec->get_section('package','');
	my $body=$mainsec->get_body();
	for (my $i=0; $i<@$body; $i++) {
	    if ($body->[$i]=~s/^(release:\s+)(\S+)//i) {
		my ($pre,$rel,$post)=($1,$2,$3);
		$rel=$self-> __increment_release($rel, $relinfo->{NEXT_RELEASE_POLICY});
		$body->[$i]=$pre.$rel.$body->[$i];
	    }
	}
	$mainsec->set_body($body);
    }

    my $new_evr=$spec->get_evr();
    $old_evr=~s/^\d+://;
    $new_evr=~s/^\d+://;
    ## HACK!!!
    if ($spec->{SOURCERPM} and -d $spec->{SOURCEDIR} and $old_evr and $new_evr) {
	foreach my $file (glob "$spec->{SOURCEDIR}/*") {
	    next unless -f $file;
	    my $basename=basename($file);
	    my $copy=$basename;
	    $copy=~s/\Q$old_evr\E/$new_evr/;
	    if ($copy ne $basename) {
		system ('cp',$file,"$spec->{SOURCEDIR}/$copy")==0 or die "cp $file $spec->{SOURCEDIR}/$copy failed!";
	    }
	}
    }

    my $packager_tag=$spec->get_section('package','')->get_tag('Packager');
    if ($relinfo->{TAG_PACKAGER} && (!$packager_tag || $relinfo->{REPLACE_TAG_PACKAGER})) {
	$spec->get_section('package','')->set_tag('Packager',$relinfo->{TAG_PACKAGER});
    }

    $spec->add_changelog(
	PACKAGER=>$relinfo->{CHANGELOG_PACKAGER},
	ENTRY=>$relinfo->{CHANGELOG},
);
}


# ------------------------

sub write_diff {
    my $self=shift;
    my $arg = {
	@_
    };

    my $spec=$self->{EDITOR};

    $arg->{DIGESTMODE}=1 unless defined $arg->{DIGESTMODE};
    $arg->{OUTPUTDIR} ||= '.';
    $arg->{SPECNAME} ||= basename($spec->{SPECFILE});

    unless ($arg->{SOURCEID}) {
	my $sourceid=$arg->{SPECNAME};
	$sourceid=~s/\.spec$//;
	if ($spec->{SOURCERPM}) {
	    $sourceid=basename($spec->{SOURCERPM});
	    $sourceid=~s/\.rpm$//;
	}
	$arg->{SOURCEID}=$sourceid;
    }

    my $diffname=$arg->{SOURCEID};
    $diffname=~s/\.src$//;

    -d $arg->{OUTPUTDIR} || mkpath([$arg->{OUTPUTDIR}]);
    $self->__mk_TMPDIR();

    my $oldspec="$self->{TMPDIR}/a/".$arg->{SPECNAME};
    my $newspec="$self->{TMPDIR}/b/".$arg->{SPECNAME};
    my $digestpatchfile="$arg->{OUTPUTDIR}/$diffname-digest.diff";
    my $solidpatchfile="$arg->{OUTPUTDIR}/$diffname.diff";
    
    my $timestamp=`LANG=C date`;
    chomp $timestamp;
    my $mailheader=qq|From repocop $timestamp
From: "$self->{PATCH_AUTHOR_NAME}" <$self->{PATCH_AUTHOR_EMAIL}>
|;
    if ($arg->{DIGESTMODE} || $arg->{BY_TEST_DIR}) {
	mkpath([dirname($oldspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
    }
    my $firststate=$spec->get_spec();
    my $beforestate=$firststate;
    my $nextstate;

    my @testnames=@{$self->{FIX_TESTNAMELIST}};
    my @pkgnames=@{$self->{FIX_PKGLIST}};
    my @messages;
    @messages=@{$self->{FIX_TESTMESSAGELIST}} if ($self->{FIX_TESTMESSAGELIST});

    # postclean scripts
    push @testnames, @postcleanscripts;

    my $i=0;
    for ($i=0;$i<@testnames;$i++) {
	if ($fixscriptmap{$testnames[$i]}) {
	    $self->_call_fixscript($testnames[$i],$pkgnames[$i]);
	    if ($arg->{DIGESTMODE} || $arg->{BY_TEST_DIR}) {
		$nextstate=$spec->get_spec();
		if ($beforestate ne $nextstate) {
		    mkpath([dirname($newspec)]);
		    &write_file($newspec, $spec->get_spec());
		    my $message=$messages[$i];
		    $message||='(message skipped)';
		    my $mailboxheader=qq|$mailheader
Subject: repocop fix for $testnames[$i]

$message
|;
		    &write_file($digestpatchfile, -e $digestpatchfile ? "\n".$mailboxheader : $mailboxheader, '>>');
		    $self->__diff($oldspec,$newspec,'>>',$digestpatchfile);
		    $self->__diff($oldspec,$newspec,'>>',"$arg->{BY_TEST_DIR}/$testnames[$i].diff") if $arg->{BY_TEST_DIR};
		    system 'mv', '-f', $newspec, $oldspec;
		    $beforestate=$nextstate;
		}
	    }
	}
    }

    $nextstate=$spec->get_spec();
    if ($firststate ne $nextstate || $arg->{OUTPUT_UNCHANGED}) {
	$self->raise_release(@{$arg->{RAISE_RELEASE}}) if $arg->{RAISE_RELEASE};
	mkpath([dirname($oldspec)]);
	mkpath([dirname($newspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
	&write_file($newspec, $spec->get_spec());
	my $mailboxheader=qq|$mailheader
Subject: repocop NMU

$self->{DEF_CHANGELOG}
|;
	&write_file($solidpatchfile, $mailboxheader, '>');
	$self->__diff($oldspec,$newspec,'>>',$solidpatchfile);
	unlink $oldspec;
	unlink $newspec;
    }
    $self->__del_TMPDIR();
}

sub DESTROY {
    my $self=shift;
    $self->__del_TMPDIR();
}

sub __diff {
    my ($self,$oldspec,$newspec,$direction, $outfile)=@_;
    -e $oldspec && -e $newspec || warn "__diff: $oldspec or $newspec does not exists!";
    my $relativeold=$oldspec;
    my $relativenew=$newspec;
    # \Q..\E here is due the bug:
    # Nested quantifiers in regex; marked by <-- HERE in m/^/home/repocop/mlt++ <-- HERE -0.2.2-alt2.9666//
    $relativeold=~s!^\Q$self->{TMPDIR}\E/!!;
    $relativenew=~s!^\Q$self->{TMPDIR}\E/!!;
    if (substr($outfile,0,1) ne '/') {
	my $curdir=`pwd`;
	chomp $curdir;
	$outfile=$curdir.'/'.$outfile;
    }
    $self->__chtmpdir_and_call("diff -u $relativeold $relativenew $direction $outfile");
}

sub __chtmpdir_and_call {
    my $self=shift;
    my $cmd=shift;
    my $curdir=`pwd`;
    chomp $curdir;
    chdir $self->{TMPDIR} || warn "can't chdir to $self->{TMPDIR}";
    system ($cmd); # && warn "__chtmpdir_and_call:in $self->{TMPDIR}: command failed: $cmd";
    chdir $curdir;
}

sub __mk_TMPDIR {
    my $self=shift;
    if (!defined $self->{TMPDIR} || -d $self->{TMPDIR}) {
	$self->{TMPDIR} = tempdir( "repocop-fix.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
	die "can't create temporary directory!" unless $self->{TMPDIR};
    } else {
	mkpath([$self->{TMPDIR}]);
    }
}

sub __del_TMPDIR {
    my $self=shift;
    if (defined $self->{TMPDIR} and -d $self->{TMPDIR}) {
	system 'chmod', '-Rf', 'u+rwX', $self->{TMPDIR};
	rmtree ([$self->{TMPDIR}]);
	undef $self->{TMPDIR};
    }
}

#--------------------------
sub load_file {
    my ($name)=@_;
    open (my $fh, '<', $name) || die "can't open $name: $!";
    local $/; # enable localized slurp mode
    my $content = <$fh>;
    close ($fh) || die "can't close $name: $!";
    return $content;
}

sub write_file {
    my ($name, $str, $mode)=@_;
    $mode ||= '>';
    open (my $fd, $mode, $name) or die "can't open $name: $!";
    print $fd $str;
    close ($fd) || die "can't close $name: $!";
}

1;
